﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Text;

using System.Drawing;
using System.Drawing.Imaging;

namespace GradientionSample
{
    public class GraphicsExtension
    {
        // 透過用のキーカラー
        public static Color KeyColor = Color.FromArgb(255, 0, 255);

        // クリッピング用画像描画デリゲート定義
        private delegate void InnderDrawFunction(Graphics g, 
            Rectangle rc, Color fillColor, Color borderColor);

        // グラデーション描画メソッド(共通処理)
        private static void InnerDrawGradient(Graphics g, Rectangle rc, 
            Color startColor, Color endColor, Color borderColor, 
            FillDirection gradientType, InnderDrawFunction func)
        {
            // クリッピング用の色の設定
            Color clippingColor = Color.Green;

            Rectangle rect = new Rectangle(0, 0, rc.Width, rc.Height);

            // グラデーションを描画するイメージの生成
            Bitmap gradientImage = new Bitmap(rc.Width, rc.Height,
                PixelFormat.Format16bppRgb565);
            Graphics gxGradient = Graphics.FromImage(gradientImage);

            // グラデーションを描画する
            GradientFill.Fill(gxGradient, rect,
                startColor, endColor, FillDirection.TopToBottom);

            // 任意の図形でグラデーションを切り抜く為のイメージの生成
            Bitmap tempImage = new Bitmap(rc.Width, rc.Height,
                PixelFormat.Format16bppRgb565);

            // クリッピング用の図形を描画するdelegateを呼ぶ
            Graphics gxTemp = Graphics.FromImage(tempImage);
            func(gxTemp, rect, clippingColor, borderColor);

            // グラデーションイメージに対して、
            // クリッピング用のイメージを、緑色の透過色を使用して上書きする
            ImageAttributes clippingAttr = new ImageAttributes();
            clippingAttr.SetColorKey(clippingColor, clippingColor);
            gxGradient.DrawImage(tempImage,
                rect, 0, 0, rect.Width, rect.Height,
                GraphicsUnit.Pixel, clippingAttr);

            // グラデーションの掛かった図形を描画する
            ImageAttributes attrib = new ImageAttributes();
            attrib.SetColorKey(KeyColor, KeyColor);
            g.DrawImage(gradientImage, rc, 0, 0, 
                gradientImage.Width, gradientImage.Height,
                GraphicsUnit.Pixel, attrib);

            // Clean up
            gxTemp.Dispose();
            tempImage.Dispose();
            gxGradient.Dispose();
            gradientImage.Dispose();
        }

        // グラデーションが掛かった円形を描画する
        public static void DrawGradientEllipse(
            Graphics g, Rectangle rc,
            Color startColor, Color endColor, Color borderColor,
            FillDirection gradientType)
        {
            InnerDrawGradient(g, rc, startColor, 
                endColor, borderColor, gradientType, 
                new InnderDrawFunction(_DrawEllipse));
        }

        // クリッピング用の図形を描画するメソッド
        private static void _DrawEllipse(Graphics g, 
            Rectangle rc, Color fillColor, Color borderColor)
        {
            // 描画用オブジェクトの取得
            SolidBrush fillBrush = new SolidBrush(fillColor);
            Pen borderPen = new Pen(borderColor);

            // 背景は透過色でクリアしておく
            g.Clear(KeyColor);

            // 円を描く
            g.FillEllipse(fillBrush, rc);

            // ボーダーを描画する
            if (borderColor != Color.Empty)
            {
                g.DrawEllipse(borderPen, rc);
            }

            // 描画用オブジェクトの解放
            fillBrush.Dispose();
            fillBrush = null;
            borderPen.Dispose();
            borderPen = null;
        }



        // グラデーションが掛かった円形を描画する
        public static void DrawGradientStar(
            Graphics g, Rectangle rc,
            Color startColor, Color endColor, Color borderColor,
            FillDirection gradientType)
        {
            InnerDrawGradient(g, rc, startColor,
                endColor, borderColor, gradientType,
                new InnderDrawFunction(_DrawStar));
        }

        // クリッピング用の図形を描画するメソッド
        private static void _DrawStar(Graphics g,
            Rectangle rc, Color fillColor, Color borderColor)
        {
            // 描画用オブジェクトの取得
            SolidBrush fillBrush = new SolidBrush(fillColor);
            Pen borderPen = new Pen(borderColor);

            // 背景は透過色でクリアしておく
            g.Clear(KeyColor);

            // 星形を描画する
            Point[] apt = new Point[5];
            for (int i = 0; i < apt.Length; i++)
            {
                double dAngle = (i * 0.8 - 0.5) * Math.PI;
               
                apt[i] = new Point((int)(rc.Width * (0.50 + 0.48 * Math.Cos(dAngle))),
                                   (int)(rc.Height * (0.50 + 0.48 * Math.Sin(dAngle))));

            }
            g.FillPolygon(fillBrush, apt);

            // ボーダーを描画する
            if (borderColor != Color.Empty)
            {
                g.DrawPolygon(borderPen, apt);
            }

            // 描画用オブジェクトの解放
            fillBrush.Dispose();
            fillBrush = null;
            borderPen.Dispose();
            borderPen = null;
        }

        public static void DrawGradientRoundRectButton(
            Graphics g, Rectangle rc,
            Color startColor, Color endColor, Color borderColor,
            FillDirection gradientType)
        {
            InnerDrawGradient(g, rc, startColor,
                endColor, borderColor, gradientType,
                new InnderDrawFunction(_DrawGradientRoundRectButton));
        }

        private static void _DrawGradientRoundRectButton(Graphics g,
            Rectangle rc, Color fillColor, Color borderColor)
        {
            // 背景は透過色でクリアしておく
            g.Clear(KeyColor);

            // 角丸矩形を描画する
            FillRoundRect(g, rc, fillColor, borderColor, 2);
        }

        public static void FillRoundRect(Graphics g,
            Rectangle rc, Color fillColor, Color borderColor, int borderWidth)
        {
            IntPtr brush = IntPtr.Zero, oldBrush = IntPtr.Zero;
            IntPtr pen = IntPtr.Zero, oldPen = IntPtr.Zero;

            IntPtr hdc = g.GetHdc();

            brush = Win32.CreateSolidBrush(Win32.RGB(fillColor));
            oldBrush = Win32.SelectObject(hdc, brush);

            if (borderColor == Color.Empty)
            {
                pen = Win32.CreatePen(0, 1, Win32.RGB(borderColor));
                oldPen = Win32.SelectObject(hdc, pen);
            }

            Win32.RoundRect(hdc, rc.Left, rc.Top, rc.Right, rc.Bottom, 20, 20);

            if (borderColor == Color.Empty)
            {
                Win32.SelectObject(hdc, oldPen);
                Win32.DeleteObject(pen);
            }
            Win32.SelectObject(hdc, oldBrush);
            Win32.DeleteObject(brush); 
            
            g.ReleaseHdc(hdc);
        }
    }
}
